/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 24/07/22.
//

#ifndef FGGL_ENTITY_ENTITY_HPP
#define FGGL_ENTITY_ENTITY_HPP

#include <cstdint>
#include <map>
#include <vector>

#include "fggl/debug/logging.hpp"
#include "fggl/vendor/entt.hpp"

namespace fggl::entity {

	using EntityID = entt::entity;
	constexpr EntityID INVALID = entt::null;

	class EntityManager {
		public:
			inline EntityID create() {
				return m_registry.create();
			}

			inline void destroy(EntityID entity) {
				m_registry.destroy(entity);
			}

			template<typename Component, typename... Args>
			inline Component &add(EntityID entity, Args &&... args) {
				return m_registry.get_or_emplace<Component>(entity, std::forward<Args>(args)...);
			}

			template<typename Component>
			Component &get(EntityID entity) {
				#ifndef NDEBUG
				if (!has<Component>(entity)) {
					debug::error("Entity {} has no component of type {}", (uint64_t) entity, debug::demangle(typeid(Component).name()));
					assert(false && "Entity was missing component - use tryGet or fix definition");
				}
				#endif
				return m_registry.get<Component>(entity);
			}

			template<typename Component>
			const Component &get(EntityID entity) const {
				return m_registry.get<Component>(entity);
			}

			template<typename Component>
			Component *tryGet(EntityID entity) {
				return m_registry.try_get<Component>(entity);
			}

			template<typename Component>
			const Component *tryGet(EntityID entity, const Component* defaultValue = nullptr) const {
				auto* comp = m_registry.try_get<Component>(entity);
				return comp == nullptr ? defaultValue : comp;
			}

			template<typename ...Components>
			auto find() const {
				return m_registry.view<Components...>();
			}

			EntityID findByName(const std::string& name) const {
				auto itr = m_names.find(name);
				if ( itr == m_names.end() ){
					return INVALID;
				}
				return m_registry.valid(itr->second) ? itr->second : INVALID;
			}

			inline void setName(const std::string& name, EntityID eid ) {
				if ( eid == INVALID ) {
					m_names.erase(name);
				} else {
					assert(m_registry.valid(eid));
					m_names[name] = eid;
				}
			}

			template<typename ...Components>
			bool has(EntityID idx) const {
				return m_registry.template all_of<Components...>(idx);
			}

			bool hasTag(EntityID entity, fggl::util::GUID tag) {
				const auto mapItr = m_tags.find( tag );
				if ( mapItr == m_tags.end() || !m_registry.valid(entity) ) {
					return false;
				}
				return std::find(mapItr->second.begin(), mapItr->second.end(), entity) != mapItr->second.end();
			}

			void addTag(const EntityID entity, const fggl::util::GUID tag) {
				assert( m_registry.valid(entity) );
				auto& tagged = m_tags[ tag ];
				tagged.push_back( entity );
			}

			void removeTag(const EntityID entity, const fggl::util::GUID tag) {
				auto mapItr = m_tags.find(tag);
				if ( mapItr == m_tags.end() ) {
					return;
				}
				std::remove_if( mapItr->second.begin(), mapItr->second.end(), [entity](auto other) { return other == entity;} );
			}

			inline std::vector<EntityID> findByTag(const char* tag) {
				return findByTag( util::make_guid_rt(tag) );
			}

			std::vector<EntityID> findByTag(const fggl::util::GUID tag){
				auto mapItr = m_tags.find(tag);
				if ( mapItr == m_tags.end() ) {
					return {};
				}
				return mapItr->second;
			}

			inline bool exists(EntityID idx) const {
				return m_registry.valid(idx);
			}

			inline bool alive(EntityID idx) const {
				return m_registry.valid(idx);
			}

		private:
			entt::registry m_registry;
			std::map<std::string, EntityID> m_names;
			std::map<fggl::util::GUID, std::vector<EntityID>> m_tags;
	};

	struct Entity {
		static Entity make(EntityManager &manager, EntityID idx) {
			return Entity{idx, manager};
		}

		EntityID id;
		EntityManager &manager;

		template<typename Component>
		Component &get() {
			return manager.get<Component>(id);
		}

		template<typename Component>
		const Component &get() const {
			return manager.get<Component>(id);
		}
	};

} // namespace fggl::entity

#endif //FGGL_ENTITY_ENTITY_HPP
